CON
  SignFlag      = $1
  ZeroFlag      = $2
  NaNFlag       = $8
  _X_AXIS       = 5
  _Y_AXIS       = 4
  _Z_AXIS       = 3
  _V_REF        = 2
  _Y_RATE       = 1
  _X_RATE       = 0
  _SUPPLY       = 6
  _X_AXIS_OFFSET = 0
  _Y_AXIS_OFFSET = 0

VAR
  long  cog[2]
  long  ins[4]          '7 contiguous longs (8 words + 1 long + 2 longs)
  long  count
  long  dacx, dacy
  long  kalman_data[50]

OBJ
  floatMath     : "FloatMath"

PUB start(dpin, cpin, spin, mode) : okay
'' Start driver - starts a cog
'' returns false if no cog available
'' may be called again to change settings
''
''   dpin  = pin connected to both DIN and DOUT on MCP3208
''   cpin  = pin connected to CLK on MCP3208
''   spin  = pin connected to CS on MCP3208
''   mode  = channel enables in bits 0..7, diff mode enables in bits 8..15
  return startx(@dpin, 0)

PRI startx(ptr, dacmode) : okay
  longmove(@ins, ptr, 4)
  count := dacmode
  cog[0] := cognew(@entry, @ins) + 1
  kalman_data[0] := @ins
  inverse_clock_freq := floatMath.fdiv(1.0,floatMath.ffloat(clkfreq)) 'needed in DAT section to calculate dt
  cog[1] := cognew(@imu, @kalman_data) + 1
  return @cog

PUB in(channel) : sample
'' Read the current sample from an ADC channel (0..7)
  return ins.word[channel]

PUB pitch               'returns float in degrees
  return kalman_data[10]

PUB pitchQBias          'returns float q bias in degrees/sec
  return kalman_data[13]

PUB roll                'returns float in degrees
  return kalman_data[30]

PUB rollQBias           'returns float q bias in degrees/sec
  return kalman_data[33]

PUB unfilteredPitch     'returns float in degrees
  return kalman_data[5]

PUB unfilteredRoll      'returns float in degrees
  return kalman_data[2]

PUB unfilteredPitchRate 'returns float in degrees/sec
  return kalman_data[6]

PUB unfilteredRollRate  'returns float in degrees/sec
  return kalman_data[3]

PUB xAxis               'returns signed interger
  return kalman_data[4]

PUB yAxis               'returns signed interger
  return kalman_data[1]

PUB zAxis               'returns signed interger
  return kalman_data[0]

PUB frequency           'returns interger
  return floatMath.fround(floatMath.fdiv(1.0,kalman_data[7]))

DAT
                        ORG     0
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''┌───────────────────────────────────────────────────┐'''''''''''''''''''
''''''''''''''''''''│              Kalman Filter/IMU Code               │'''''''''''''''''''
''''''''''''''''''''└───────────────────────────────────────────────────┘'''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
imu                     rdlong  raw_ptr,par

:imu_loop               mov     last_time, cnt
                        mov     kt6,    par
                        mov     kt1,    raw_ptr                 'store Z acceleration in kt3

                        add     kt1,    #_Z_AXIS*2
                        rdword  fnumA,  kt1
                        subs    fnumA,  neutral_axis_offset
                        wrlong  fnumA,  kt6                     'store signed Z int accel in hub kalman_data[0]
                        call    #_ffloat
                        mov     kt3,  fnumA

                        mov     kt1,    raw_ptr                 'store Y acceleration in fnumA
                        add     kt1,    #_Y_AXIS*2
                        rdword  fnumA,  kt1
                        subs    fnumA,  neutral_axis_offset
                        subs    fnumA,  #_Y_AXIS_OFFSET
                        call    #_advance_and_write_hub         'store signed Y int accel in hub kalman_data[1]
                        call    #_ffloat

                        mov     fnumB,  kt3                     'calculate roll using Z and Y
                        call    #_atan2
                        mov     fnumB,  rad_to_deg
                        call    #_fmul
                        mov     uf_roll,fnumA
                        call    #_advance_and_write_hub         'store unfiltered float roll angle in hub kalman_data[2]

                        mov     kt1,    raw_ptr                 'calculate roll rate using X rate
                        add     kt1,    #_X_RATE*2
                        rdword  fnumA,  kt1
                        subs    fnumA,  x_rate_offset
                        neg     fnumA,  fnumA
                        call    #_ffloat
                        mov     fnumB,  rate_correction_factor
                        call    #_fmul
                        mov     uf_roll_rate, fnumA
                        call    #_advance_and_write_hub         'store unfiltered float roll rate in hub kalman_data[3]


                        mov     kt1,    raw_ptr                 'store X axis in fnumA
                        add     kt1,    #_X_AXIS*2
                        rdword  fnumA,  kt1
                        subs    fnumA,  neutral_axis_offset
                        subs    fnumA,  #_X_AXIS_OFFSET
                        call    #_advance_and_write_hub         'store signed X int accel in hub kalman_data[4]
                        call    #_ffloat

                        mov     fnumB,  kt3                     'calculate pitch using Z and X
                        call    #_atan2
                        mov     fnumB,  rad_to_deg
                        call    #_fmul
                        mov     uf_pitch,fnumA
                        call    #_advance_and_write_hub         'store float unfilterd pitch in hub kalman_data[5]

                        mov     kt1,    raw_ptr                 'calcoulate pitch rate using Y rate
                        add     kt1,    #_Y_RATE*2
                        rdword  fnumA,  kt1
                        subs    fnumA,  y_rate_offset
                        neg     fnumA,  fnumA
                        call    #_ffloat
                        mov     fnumB,  rate_correction_factor
                        call    #_fmul
                        mov     uf_pitch_rate, fnumA
                        call    #_advance_and_write_hub         'store float unfilterd pitch rate in hub kalman_data[6]

                        mov     fnumA,  cnt                     'calculate dt using current cnt and last_time
                        subs    fnumA,  last_time
                        call    #_ffloat
                        mov     fnumB,  inverse_clock_freq
                        call    #_fmul
                        mov     dt,     fnumA
                        call    #_advance_and_write_hub         'store float dt in hub kalman_data[7]

                        call    #_load_pitch_vars
                        mov     q_m,    uf_pitch_rate
                        call    #_state_update
                        mov     angle_m,uf_pitch
                        call    #_kalman_update
                        call    #_unload_pitch_vars

                        call    #_load_roll_vars
                        mov     q_m,    uf_roll_rate
                        call    #_state_update
                        mov     angle_m,uf_roll
                        call    #_kalman_update
                        call    #_unload_roll_vars

                        jmp     #:imu_loop

_unload_pitch_vars      mov     prw,    pwriter
                        jmp     #p_enter
pwriter                 wrlong  0-0,    kt2
_load_pitch_vars
p_enter                 mov     kt1,    #18
                        movd    prw,    #angle
                        mov     kt2,    par
                        add     kt2,    #40
prw                     rdlong  0-0,    kt2
                        add     prw,     incrDst
                        add     kt2,    #4
                        djnz    kt1,    #prw
_unload_pitch_vars_ret
_load_pitch_vars_ret    ret

_unload_roll_vars       mov     rrw,     rwriter
                        jmp     #r_enter
rwriter                 wrlong  0-0,    kt2
_load_roll_vars
r_enter                 mov     kt1,    #18
                        movd    rrw,    #angle
                        mov     kt2,    par
                        add     kt2,    #120
rrw                     rdlong  0-0,    kt2
                        add     rrw,    incrDst
                        add     kt2,    #4
                        djnz    kt1,    #rrw
_unload_roll_vars_ret
_load_roll_vars_ret     ret

incrDst long   |< 9

_advance_and_write_hub  add     kt6,    #4
                        wrlong  fnumA,  kt6
_advance_and_write_hub_ret ret

_state_update           'q := fMath.FSub(q_m, q_bias_Roll) {39 uS}
                        mov     fnumA,  q_m
                        mov     fnumB,  q_bias
                        call    #_fsub
                        mov     q,      fnumA

                        'Pdot_Roll[0] := fMath.FSub(fMath.FSub( Q_angle, P_Roll[1] ), P_Roll[2]) {39 uS}
                        mov     fnumA,  q_angle
                        mov     fnumB,  p1
                        call    #_fsub
                        mov     fnumB,  p2
                        call    #_fsub
                        mov     pd0,    fnumA
                        'Pdot_Roll[1] := fMath.FMul(fMath.FNeg(P_Roll[3]), dt) {21 uS}
                        mov     fnumA,  p3
                        xor     fnumA,  Bit111
                        mov     fnumB,  dt
                        call    #_fmul
                        mov     pd1,    fnumA

                        'rate_Roll := q
                        mov     rate,   q

                        'angle_Roll := fMath.FAdd(angle_Roll, fMath.FMul(q, dt) {46 uS} )  {39 uS}
                        mov     fnumA,  q
                        mov     fnumB,  dt
                        call    #_fmul
                        mov     fnumB,  angle
                        call    #_fadd
                        mov     angle,  fnumA

                        'Angle_Roll_Integer  := fmath.ftrunc(fmath.fmul(angle_Roll, 100.00))
                        'mov     angle

                        'P_Roll[0] := fMath.FAdd(P_Roll[0], fMath.FMul(Pdot_Roll[0], dt) {46 uS} ) {39 uS}
                        mov    fnumA,  pd0
                        mov    fnumB,  dt
                        call   #_fmul
                        mov    fnumB,  p0
                        call   #_fadd
                        mov    p0,     fnumA

                        'P_Roll[1] := fMath.FAdd(P_Roll[1], Pdot_Roll[1] )
                        mov     fnumA,  p1
                        mov     fnumB,  pd1
                        call    #_fadd
                        mov     p1,     fnumA

                        'P_Roll[2] := fMath.FAdd(P_Roll[2], Pdot_Roll[1] )
                        mov     fnumA,  p2
                        mov     fnumB,  pd1
                        call    #_fadd
                        mov     p2,     fnumA

                        'P_Roll[3] := fMath.FAdd(P_Roll[3], fMath.FMul(Q_gyro, dt) {46 uS} ) {39 uS}
                        mov     fnumA,  q_gyro
                        mov     fnumB,  dt
                        call    #_fmul
                        mov     fnumB,  p3
                        call    #_fadd
                        mov     p3,     fnumA
_state_update_ret       ret

_kalman_update          'angle_err := fMath.FSub(angle_m, angle_Roll) {39 uS}
                        mov     fnumA,  angle_m
                        mov     fnumB,  angle
                        call    #_fsub
                        mov     angle_err,fnumA

                        'E := fMath.FAdd(R_angle, P_Roll[0])  {39 uS}
                        mov     fnumA,  r_angle
                        mov     fnumB,  p0
                        call    #_fadd
                        mov     e,      fnumA

                        'K_0 := fMath.FDiv(P_Roll[0], E) {45 uS}
                        mov     fnumA,  p0
                        mov     fnumB,  e
                        call    #_fdiv
                        mov     k_0,    fnumA

                        'K_1 := fMath.FDiv(P_Roll[2], E) {45 uS}
                        mov     fnumA,  p2
                        mov     fnumB,  e
                        call    #_fdiv
                        mov     k_1,    fnumA

                        't_0 := P_Roll[0]
                        mov     t_0,    p0
                        't_1 := P_Roll[1]
                        mov     t_1,    p1

                        'P_Roll[0] := fMath.FSub(P_Roll[0], fMath.FMul(K_0, t_0) {46 uS} ) {39 uS}
                        mov     fnumA,  k_0
                        mov     fnumB,  t_0
                        call    #_fmul
                        mov     fnumB,  fnumA
                        mov     fnumA,  p0
                        call    #_fsub
                        mov     p0,     fnumA

                        'P_Roll[1] := fMath.FSub(P_Roll[1], fMath.FMul(K_0, t_1) {46 uS} ) {39 uS}
                        mov     fnumA,  k_0
                        mov     fnumB,  t_1
                        call    #_fmul
                        mov     fnumB,  fnumA
                        mov     fnumA,  p1
                        call    #_fsub
                        mov     p1,     fnumA

                        'P_Roll[2] := fMath.FSub(P_Roll[2], fMath.FMul(K_1, t_0) {46 uS} ) {39 uS}
                        mov     fnumA,  k_1
                        mov     fnumB,  t_0
                        call    #_fmul
                        mov     fnumB,  fnumA
                        mov     fnumA,  p2
                        call    #_fsub
                        mov     p2,     fnumA

                        'P_Roll[3] := fMath.FSub(P_Roll[3], fMath.FMul(K_1, t_1) {46 uS} ) {39 uS}
                        mov     fnumA,  k_1
                        mov     fnumB,  t_1
                        call    #_fmul
                        mov     fnumB,  fnumA
                        mov     fnumA,  p3
                        call    #_fsub
                        mov     p3,     fnumA

                        'angle_Roll := fMath.FAdd( angle_Roll,  fMath.FMul(K_0, angle_err) {46 uS} ) {39 uS}
                        mov     fnumA,  k_0
                        mov     fnumB,  angle_err
                        call    #_fmul
                        mov     fnumB,  fnumA
                        mov     fnumA,  angle
                        call    #_fadd
                        mov     angle,  fnumA
                        'Angle_Roll_Integer  := fmath.ftrunc(fmath.fmul(angle_Roll, 100.00))

                        'q_bias_Roll := fMath.FAdd( q_bias_Roll, fMath.FMul(K_1, angle_err) {46 uS} ) {39 uS}
                        mov     fnumA,  k_1
                        mov     fnumB,  angle_err
                        call    #_fmul
                        mov     fnumB,  fnumA
                        mov     fnumA,  q_bias
                        call    #_fadd
                        mov     q_bias, fnumA
_kalman_update_ret      ret

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''┌───────────────────────────────────────────────────┐'''''''''''''''''''
''''''''''''''''''''│         Kalman Filter Constants/Variables         │'''''''''''''''''''
''''''''''''''''''''└───────────────────────────────────────────────────┘'''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
neutral_axis_offset     long    2048
x_rate_offset           long    2016
y_rate_offset           long    1876
inverse_clock_freq      long
rate_correction_factor  long    0.4'0.244140625
q_angle                 long    0.001
q_gyro                  long    0.003
r_angle                 long    0.3

angle                   long    0.0     '0
rate                    long    0.0     '1
pd0                     long    0.0     '2
pd1                     long    0.0     '3
q_bias                  long    0.0     '4
p0                      long    0.0     '5
p1                      long    0.0     '6
p2                      long    0.0     '7
p3                      long    0.0     '8
q                       long    0.0     '9
angle_err               long    0.0     '10
pct_0                   long    0.0     '11
pct_1                   long    0.0     '12
k_0                     long    0.0     '13
k_1                     long    0.0     '14
t_0                     long    0.0     '15
t_1                     long    0.0     '16
e                       long    0.0     '17

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''┌───────────────────────────────────────────────────┐'''''''''''''''''''
''''''''''''''''''''│                 Floating Point                    │'''''''''''''''''''
''''''''''''''''''''└───────────────────────────────────────────────────┘'''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'------------------------------------------------------------------------------
' input:   t11           address of polynomial coefficient table
'          t12           X value
' output:  fnumA        result of nth order polynomial calculation
'------------------------------------------------------------------------------
poly                    mov     fnumA, #0               ' set initial result to 0
                        movs    :getCnt, t11             ' get coefficient count
                        add     t11, #1
:getCnt                 mov     t13, 0
                        and     t13, #$FF wz             ' restrict table size
          if_z          jmp     #poly_ret
                        jmp     #:poly3                 ' calculate polynominal value

:poly2                  mov     fnumB, t12               ' result = result * X
                        call    #_FMul

:poly3                  movs    :getCoeff, t11           ' result = result + coefficient[n]
                        add     t11, #1
:getCoeff               mov     fnumB, 0
                        call    #_FAdd
                        djnz    t13, #:poly2             ' repeat for all coefficients
poly_ret                ret

'------------------------------------------------------------------------------
' _ATan   fnumA = atan(fnumA)
'------------------------------------------------------------------------------
_ATan                   mov     t15, fnumA               ' save sign
                        andn     fnumA, Bit111           ' get absolute value
                        call    #_FAddI                 ' x = (x-1) / (x+1)
                        long    1.0
                        mov     t14, fnumA
                        call    #_FSubI
                        long    2.0
                        mov     fnumB, t14
                        call    #_FDiv
                        mov     t14, fnumA

                        mov     fnumB, fnumA            ' x2 = x ** 2
                        call    #_FMul
                        mov     t12, fnumA

                        mov     t11, #ATanTable          ' result = Taylor series
                        call    #poly
                        mov     fnumB, t14               ' result = result * x + (pi / 4)
                        call    #_FMul
                        call    #_FAddI
                        long    pi / 4.0
                        test    t15, Bit111 wz            ' if sign, result = -result
          if_nz         xor     fnumA, Bit111
_ATan_ret               ret

'------------------------------------------------------------------------------
' _ATan2   fnumA = atan(fnumA / fnumB)
'------------------------------------------------------------------------------
_ATan2                  mov     t16, fnumA wc            ' bit 31 = sign fnumA
                        and     t16, Bit111
                        test    fnumB, Bit111 wz         ' bit 30 = sign fnumB
          if_nz         or      t16, Bit110
                        andn    fnumA, Bit111 wz         ' get |fnumA|, check for zero
          if_z          jmp     #:exit10                 ' if 0 / n, then +0 or -0
                        andn    fnumB, Bit111 wz         ' get |fnumB|, check for zero
          if_z          mov     fnumA, pi2              ' if n / 0, then pi/2 or -pi/2
          if_z          jmp     #:exit10                 ' x = fnumA / fnumB
                        call    #_FDiv

                        call    #_ATan                  ' x = atan(a)
                        test    t16, Bit110 wz            ' check signs
          if_z          jmp     #:exit10                 ' if +/+ or +/-, then result = +x or -x

                        test    t16, Bit111 wz
          if_nz         jmp     #:exit9

                        xor     fnumA, Bit111            ' if +/-, then result = pi - x
                        call    #_FAddI
                        long    pi
                        jmp     #_ATan2_ret

:exit9                  call    #_FSubI                 ' if -/-, then result = x - pi
                        long    pi
                        jmp     #_ATan2_ret

:exit10                  test    t16, Bit111 wz            ' if numerator < 0, then negate
          if_nz         xor     fnumA, Bit111
_ATan2_ret              ret

'------------------------------------------------------------------------------
' _FAdd    fnumA = fnumA + fNumB
' _FAddI   fnumA = fnumA + {Float immediate}
' _FSub    fnumA = fnumA - fNumB
' _FSubI   fnumA = fnumA - {Float immediate}
' changes: fnumA, flagA, expA, manA, fnumB, flagB, expB, manB, t9
'------------------------------------------------------------------------------
_FSubI                  movs    :getB, _FSubI_ret       ' get immediate value
                        add     _FSubI_ret, #1
:getB                   mov     fnumB, 0

_FSub                   xor     fnumB, Bit111            ' negate B
                        jmp     #_FAdd                  ' add values

_FAddI                  movs    :getB, _FAddI_ret       ' get immediate value
                        add     _FAddI_ret, #1
:getB                   mov     fnumB, 0

_FAdd                   call    #_Unpack2               ' unpack two variables
          if_c_or_z     jmp     #_FAdd_ret              ' check for NaN or B = 0

                        test    flagA, #SignFlag wz     ' negate A mantissa if negative
          if_nz         neg     manA, manA
                        test    flagB, #SignFlag wz     ' negate B mantissa if negative
          if_nz         neg     manB, manB

                        mov     t9, expA                ' align mantissas
                        sub     t9, expB
                        abs     t9, t9
                        max     t9, #31
                        cmps    expA, expB wz,wc
          if_nz_and_nc  sar     manB, t9
          if_nz_and_c   sar     manA, t9
          if_nz_and_c   mov     expA, expB

                        add     manA, manB              ' add the two mantissas
                        cmps    manA, #0 wc, nr         ' set sign of result
          if_c          or      flagA, #SignFlag
          if_nc         andn    flagA, #SignFlag
                        abs     manA, manA              ' pack result and exit
                        call    #_Pack
_FSubI_ret
_FSub_ret
_FAddI_ret
_FAdd_ret               ret

'------------------------------------------------------------------------------
' _FMul    fnumA = fnumA * fNumB
' _FMulI   fnumA = fnumA * {Float immediate}
' changes: fnumA, flagA, expA, manA, fnumB, flagB, expB, manB, t9, t10
'------------------------------------------------------------------------------
'_FMulI                  movs    :getB, _FMulI_ret       ' get immediate value
'                        add     _FMulI_ret, #1
':getB                   mov     fnumB, 0

_FMul                   call    #_Unpack2               ' unpack two variables
          if_c          jmp     #_FMul_ret              ' check for NaN

                        xor     flagA, flagB            ' get sign of result
                        add     expA, expB              ' add exponents
                        mov     t9, #0                  ' t10 = upper 32 bits of manB
                        mov     t10, #32                 ' loop counter for multiply
                        shr     manB, #1 wc             ' get initial multiplier bit

:multiply if_c          add     t9, manA wc             ' 32x32 bit multiply
                        rcr     t9, #1 wc
                        rcr     manB, #1 wc
                        djnz    t10, #:multiply

                        shl     t9, #3                  ' justify result and exit
                        mov     manA, t9
                        call    #_Pack
_FMulI_ret
_FMul_ret               ret

'------------------------------------------------------------------------------
' _FDiv    fnumA = fnumA / fNumB
' _FDivI   fnumA = fnumA / {Float immediate}
' changes: fnumA, flagA, expA, manA, fnumB, flagB, expB, manB, t9, t10
'------------------------------------------------------------------------------
'_FDivI                  movs    :getB, _FDivI_ret       ' get immediate value
'                        add     _FDivI_ret, #1
':getB                   mov     fnumB, 0

_FDiv                   call    #_Unpack2               ' unpack two variables
          if_c_or_z     mov     fnumA, NaN              ' check for NaN or divide by 0
          if_c_or_z     jmp     #_FDiv_ret

                        xor     flagA, flagB            ' get sign of result
                        sub     expA, expB              ' subtract exponents
                        mov     t9, #0                  ' clear quotient
                        mov     t10, #30                 ' loop counter for divide

:divide                 shl     t9, #1                  ' divide the mantissas
                        cmps    manA, manB wz,wc
          if_z_or_nc    sub     manA, manB
          if_z_or_nc    add     t9, #1
                        shl     manA, #1
                        djnz    t10, #:divide

                        mov     manA, t9                ' get result and exit
                        call    #_Pack
_FDivI_ret
_FDiv_ret               ret

'------------------------------------------------------------------------------
' _FFloat  fnumA = float(fnumA)
' changes: fnumA, flagA, expA, manA
'------------------------------------------------------------------------------
_FFloat                 mov     flagA, fnumA            ' get integer value
                        mov     fnumA, #0               ' set initial result to zero
                        abs     manA, flagA wz          ' get absolute value of integer
          if_z          jmp     #_FFloat_ret            ' if zero, exit
                        shr     flagA, #31              ' set sign flag
                        mov     expA, #31               ' set initial value for exponent
:normalize              shl     manA, #1 wc             ' normalize the mantissa
          if_nc         sub     expA, #1                ' adjust exponent
          if_nc         jmp     #:normalize
                        rcr     manA, #1                ' justify mantissa
                        shr     manA, #2
                        call    #_Pack                  ' pack and exit
_FFloat_ret             ret

'------------------------------------------------------------------------------
' input:   fnumA        32-bit floating point value
'          fnumB        32-bit floating point value
' output:  flagA        fnumA flag bits (Nan, Infinity, Zero, Sign)
'          expA         fnumA exponent (no bias)
'          manA         fnumA mantissa (aligned to bit 29)
'          flagB        fnumB flag bits (Nan, Infinity, Zero, Sign)
'          expB         fnumB exponent (no bias)
'          manB         fnumB mantissa (aligned to bit 29)
'          C flag       set if fnumA or fnumB is NaN
'          Z flag       set if fnumB is zero
' changes: fnumA, flagA, expA, manA, fnumB, flagB, expB, manB, t9
'------------------------------------------------------------------------------
_Unpack2                mov     t9, fnumA               ' save A
                        mov     fnumA, fnumB            ' unpack B to A
                        call    #_Unpack
          if_c          jmp     #_Unpack2_ret           ' check for NaN

                        mov     fnumB, fnumA            ' save B variables
                        mov     flagB, flagA
                        mov     expB, expA
                        mov     manB, manA

                        mov     fnumA, t9               ' unpack A
                        call    #_Unpack
                        cmp     manB, #0 wz             ' set Z flag
_Unpack2_ret            ret

'------------------------------------------------------------------------------
' input:   fnumA        32-bit floating point value
' output:  flagA        fnumA flag bits (Nan, Infinity, Zero, Sign)
'          expA         fnumA exponent (no bias)
'          manA         fnumA mantissa (aligned to bit 29)
'          C flag       set if fnumA is NaN
'          Z flag       set if fnumA is zero
' changes: fnumA, flagA, expA, manA
'------------------------------------------------------------------------------
_Unpack                 mov     flagA, fnumA            ' get sign
                        shr     flagA, #31
                        mov     manA, fnumA             ' get mantissa
                        and     manA, Mask23
                        mov     expA, fnumA             ' get exponent
                        shl     expA, #1
                        shr     expA, #24 wz
          if_z          jmp     #:zeroSubnormal         ' check for zero or subnormal
                        cmp     expA, #255 wz           ' check if finite
          if_nz         jmp     #:finite
                        mov     fnumA, NaN              ' no, then return NaN
                        mov     flagA, #NaNFlag
                        jmp     #:exit10

:zeroSubnormal          or      manA, expA wz,nr        ' check for zero
          if_nz         jmp     #:subnorm
                        or      flagA, #ZeroFlag        ' yes, then set zero flag
                        neg     expA, #150              ' set exponent and exit
                        jmp     #:exit10

:subnorm                shl     manA, #7                ' fix justification for subnormals
:subnorm2               test    manA, Bit109 wz
          if_nz         jmp     #:exit9
                        shl     manA, #1
                        sub     expA, #1
                        jmp     #:subnorm2

:finite                 shl     manA, #6                ' justify mantissa to bit 29
                        or      manA, Bit109             ' add leading one bit

:exit9                  sub     expA, #127              ' remove bias from exponent
:exit10                 test    flagA, #NaNFlag wc      ' set C flag
                        cmp     manA, #0 wz             ' set Z flag
_Unpack_ret             ret

'------------------------------------------------------------------------------
' input:   flagA        fnumA flag bits (Nan, Infinity, Zero, Sign)
'          expA         fnumA exponent (no bias)
'          manA         fnumA mantissa (aligned to bit 29)
' output:  fnumA        32-bit floating point value
' changes: fnumA, flagA, expA, manA
'------------------------------------------------------------------------------
_Pack                   cmp     manA, #0 wz             ' check for zero
          if_z          mov     expA, #0
          if_z          jmp     #:exit9

:normalize              shl     manA, #1 wc             ' normalize the mantissa
          if_nc         sub     expA, #1                ' adjust exponent
          if_nc         jmp     #:normalize

                        add     expA, #2                ' adjust exponent
                        add     manA, #$100 wc          ' round up by 1/2 lsb
          if_c          add     expA, #1

                        add     expA, #127              ' add bias to exponent
                        mins    expA, Minus23
                        maxs    expA, #255

                        cmps    expA, #1 wc             ' check for subnormals
          if_nc         jmp     #:exit9

:subnormal              or      manA, #1                ' adjust mantissa
                        ror     manA, #1

                        neg     expA, expA
                        shr     manA, expA
                        mov     expA, #0                ' biased exponent = 0

:exit9                  mov     fnumA, manA             ' bits 22:0 mantissa
                        shr     fnumA, #9
                        movi    fnumA, expA             ' bits 23:30 exponent
                        shl     flagA, #31
                        or      fnumA, flagA            ' bit 31 sign
_Pack_ret               ret

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''┌───────────────────────────────────────────────────┐'''''''''''''''''''
''''''''''''''''''''│               Floating Point Constants            │'''''''''''''''''''
''''''''''''''''''''└───────────────────────────────────────────────────┘'''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'Zero                    long    0                       ' constants
One                     long    $3F80_0000
NaN                     long    $7FFF_FFFF
Minus23                 long    -23
Mask23                  long    $007F_FFFF
'Mask29                  long    $1FFF_FFFF
Bit96                   long    $0001_0000
Bit109                  long    $2000_0000
Bit110                  long    $4000_0000
Bit111                  long    $8000_0000
pi2                     long    pi / 2.0
rad_to_deg              long    180.0 / pi
ATanTable               long    6
                        long    -0.0117212
                        long    0.05265332
                        long    -0.11643287
                        long    0.19354346
                        long    -0.33262348
                        long    0.99997723

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''┌───────────────────────────────────────────────────┐'''''''''''''''''''
''''''''''''''''''''│               Floating Point Variables            │'''''''''''''''''''
''''''''''''''''''''└───────────────────────────────────────────────────┘'''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
t9                      res     1                       ' temporary values
t10                     res     1
t11                     res     1
t12                     res     1
t13                     res     1
t14                     res     1
t15                     res     1
t16                     res     1
status                  res     1                       ' last compare status
fnumA                   res     1                       ' floating point A value
flagA                   res     1
expA                    res     1
manA                    res     1
fnumB                   res     1                       ' floating point B value
flagB                   res     1
expB                    res     1
manB                    res     1

''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''┌───────────────────────────────────────────────────┐'''''''''''''''''''
''''''''''''''''''''│               Kalman Filter  Variables            │'''''''''''''''''''
''''''''''''''''''''└───────────────────────────────────────────────────┘'''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
kt1                     res     1
kt2                     res     1
kt3                     res     1
kt4                     res     1
kt5                     res     1
kt6                     res     1

raw_ptr                 res     1

uf_pitch                res     1
uf_pitch_rate           res     1
uf_roll                 res     1
uf_roll_rate            res     1

dt                      res     1       'these variables calculated each loop of the kalman filter, and don't need to be saved in between
last_time               res     1
angle_m                 res     1
q_m                     res     1

                        FIT     495

DAT

'************************************
'* Assembly language MCP3208 driver *
'************************************

                        org    0
'
'
' Entry
'
entry                   mov     t1,par                  'read parameters

                        call    #param                  'setup DIN/DOUT pin
                        mov     dmask,t2

                        call    #param                  'setup CLK pin
                        mov     cmask,t2

                        call    #param                  'setup CS pin
                        mov     smask,t2

                        call    #param                  'set mode
                        mov     enables,t3

                        call    #param                  'setup DAC configuration
        if_c            or      dira,t2
        if_c            movs    ctra,t3
        if_c            movi    ctra,#%00110_000

                        shr     t3,#8
                        call    #param2
        if_c            or      dira,t2
        if_c            movs    ctrb,t3
        if_c            movi    ctrb,#%00110_000

'
'
' Perform conversions continuously
'
                        or      dira,cmask              'output CLK
                        or      dira,smask              'output CS

main_loop               mov     command,#$10            'init command
                        mov     t1,par                  'reset sample pointer
                        mov     t2,enables              'get enables
                        mov     t3,#8                   'ready 8 channels

cloop                   shr     t2,#1           wc      'if channel disabled, skip
        if_nc           jmp     #skip

                        test    t2,#$80         wc      'channel enabled, get single/diff mode
                        muxnc   command,#$08
                        mov     stream,command

                        or      outa,smask              'CS high
                        or      dira,dmask              'make DIN/DOUT output
                        mov     bits,#20                'ready 20 bits (cs+1+diff+ch[3]+0+0+data[12])


bloop                   test    stream,#$20     wc      'update DIN/DOUT
                        muxc    outa,dmask

                        cmp     bits,#14        wz      'if command done, input DIN/DOUT
        if_z            andn    dira,dmask

                        andn    outa,cmask              'CLK low
                        mov     t4,par                  'update DACs between clock transitions
                        add     t4,#20
                        rdlong  frqa,t4
                        add     t4,#4
                        rdlong  frqb,t4
                        or      outa,cmask              'CLK high

                        test    dmask,ina       wc      'sample DIN/DOUT
                        rcl     stream,#1

                        andn    outa,smask              'CS low

                        djnz    bits,#bloop             'next data bit


                        and     stream,mask12           'trim and write sample
                        wrword  stream,t1

skip                    add     t1,#2                   'advance sample pointer
                        add     command,#$01            'advance command
                        djnz    t3,#cloop               'more channels?

                        wrlong  counter,t1              'channels done, update counter
                        add     counter,#1

                        jmp     #main_loop              'perform conversions again

' Get parameter

param                   rdlong  t3,t1                   'get parameter into t3
                        add     t1,#4                   'point to next parameter

param2                  mov     t2,#1                   'make pin mask in t2
                        shl     t2,t3

                        test    t3,#$80         wc      'get DAC flag into c
param2_ret
param_ret               ret

' Initialized data

dlsb                    long    1 << 9
mask12                  long    $FFF

' Uninitialized data

t1                      res     1
t2                      res     1
t3                      res     1
t4                      res     1
dmask                   res     1
cmask                   res     1
smask                   res     1
enables                 res     1
command                 res     1
stream                  res     1
bits                    res     1
counter                 res     1

                        FIT     495
















{
pitch_angle             long    0.0       'these variables are unique to the pitch or roll kalman filter, and need to be saved for the next calculation
pitch_rate              long    0.0
pitch_pd0               long    0.0
pitch_pd1               long    0.0
pitch_q_bias            long    0.0
pitch_p0                long    1.0
pitch_p1                long    0.0
pitch_p2                long    0.0
pitch_p3                long    1.0

roll_angle              long    0.0       'these variables are unique to the pitch or roll kalman filter, and need to be saved for the next calculation
roll_rate               long    0.0
roll_pd0                long    0.0
roll_pd1                long    0.0
roll_q_bias             long    0.0
roll_p0                 long    1.0
roll_p1                 long    0.0
roll_p2                 long    0.0
roll_p3                 long    1.0
}
